import datetime

from django.db.models import Q
from django.utils import timezone

import stripe

from .. import hooks, models, utils


def cancel(subscription, at_period_end=True):
    """
    Cancels a subscription

    Args:
        subscription: the subscription to cancel
        at_period_end: True to cancel at the end of the period, otherwise cancels immediately
    """
    sub = stripe.Subscription(
        subscription.stripe_id,
        stripe_account=subscription.stripe_account_stripe_id,
    ).delete(
        at_period_end=at_period_end,
    )
    return sync_subscription_from_stripe_data(subscription.customer, sub)


def create(customer, plan, quantity=None, trial_days=None, token=None, coupon=None, tax_percent=None):
    """
    Creates a subscription for the given customer

    Args:
        customer: the customer to create the subscription for
        plan: the plan to subscribe to
        quantity: if provided, the number to subscribe to
        trial_days: if provided, the number of days to trial before starting
        token: if provided, a token from Stripe.js that will be used as the
               payment source for the subscription and set as the default
               source for the customer, otherwise the current default source
               will be used
        coupon: if provided, a coupon to apply towards the subscription
        tax_percent: if provided, add percentage as tax

    Returns:
        the pinax.stripe.models.Subscription object (created or updated)
    """
    quantity = hooks.hookset.adjust_subscription_quantity(customer=customer, plan=plan, quantity=quantity)

    subscription_params = {}
    if trial_days:
        subscription_params["trial_end"] = datetime.datetime.utcnow() + datetime.timedelta(days=trial_days)
    if token:
        subscription_params["source"] = token

    subscription_params["stripe_account"] = customer.stripe_account_stripe_id
    subscription_params["customer"] = customer.stripe_id
    subscription_params["plan"] = plan
    subscription_params["quantity"] = quantity
    subscription_params["coupon"] = coupon
    subscription_params["tax_percent"] = tax_percent
    resp = stripe.Subscription.create(**subscription_params)

    return sync_subscription_from_stripe_data(customer, resp)


def has_active_subscription(customer):
    """
    Checks if the given customer has an active subscription

    Args:
        customer: the customer to check

    Returns:
        True, if there is an active subscription, otherwise False
    """
    return models.Subscription.objects.filter(
        customer=customer
    ).filter(
        Q(ended_at__isnull=True) | Q(ended_at__gt=timezone.now())
    ).exists()


def is_period_current(subscription):
    """
    Tests if the provided subscription object for the current period

    Args:
        subscription: a pinax.stripe.models.Subscription object to test
    """
    return subscription.current_period_end > timezone.now()


def is_status_current(subscription):
    """
    Tests if the provided subscription object has a status that means current

    Args:
        subscription: a pinax.stripe.models.Subscription object to test
    """
    return subscription.status in subscription.STATUS_CURRENT


def is_valid(subscription):
    """
    Tests if the provided subscription object is valid

    Args:
        subscription: a pinax.stripe.models.Subscription object to test
    """
    if not is_status_current(subscription):
        return False

    if subscription.cancel_at_period_end and not is_period_current(subscription):
        return False

    return True


def retrieve(customer, sub_id):
    """
    Retrieve a subscription object from Stripe's API

    Args:
        customer: a legacy argument, we check that the given
            subscription belongs to the given customer
        sub_id: the Stripe ID of the subscription you are fetching

    Returns:
        the data for a subscription object from the Stripe API
    """
    if not sub_id:
        return
    subscription = stripe.Subscription.retrieve(sub_id, stripe_account=customer.stripe_account_stripe_id)
    if subscription and subscription.customer != customer.stripe_id:
        return
    return subscription


def sync_subscription_from_stripe_data(customer, subscription):
    """
    Synchronizes data from the Stripe API for a subscription

    Args:
        customer: the customer who's subscription you are syncronizing
        subscription: data from the Stripe API representing a subscription

    Returns:
        the pinax.stripe.models.Subscription object (created or updated)
    """
    defaults = dict(
        customer=customer,
        application_fee_percent=subscription["application_fee_percent"],
        cancel_at_period_end=subscription["cancel_at_period_end"],
        canceled_at=utils.convert_tstamp(subscription["canceled_at"]),
        current_period_start=utils.convert_tstamp(subscription["current_period_start"]),
        current_period_end=utils.convert_tstamp(subscription["current_period_end"]),
        ended_at=utils.convert_tstamp(subscription["ended_at"]),
        plan=models.Plan.objects.get(stripe_id=subscription["plan"]["id"]),
        quantity=subscription["quantity"],
        start=utils.convert_tstamp(subscription["start"]),
        status=subscription["status"],
        trial_start=utils.convert_tstamp(subscription["trial_start"]) if subscription["trial_start"] else None,
        trial_end=utils.convert_tstamp(subscription["trial_end"]) if subscription["trial_end"] else None
    )
    sub, created = models.Subscription.objects.get_or_create(
        stripe_id=subscription["id"],
        defaults=defaults
    )
    sub = utils.update_with_defaults(sub, defaults, created)
    return sub


def update(subscription, plan=None, quantity=None, prorate=True, coupon=None, charge_immediately=False):
    """
    Updates a subscription

    Args:
        subscription: the subscription to update
        plan: optionally, the plan to change the subscription to
        quantity: optionally, the quantity of the subscription to change
        prorate: optionally, if the subscription should be prorated or not
        coupon: optionally, a coupon to apply to the subscription
        charge_immediately: optionally, whether or not to charge immediately
    """
    stripe_subscription = subscription.stripe_subscription
    if plan:
        stripe_subscription.plan = plan
    if quantity:
        stripe_subscription.quantity = quantity
    if not prorate:
        stripe_subscription.prorate = False
    if coupon:
        stripe_subscription.coupon = coupon
    if charge_immediately:
        if stripe_subscription.trial_end is not None and utils.convert_tstamp(stripe_subscription.trial_end) > timezone.now():
            stripe_subscription.trial_end = "now"
    sub = stripe_subscription.save()
    customer = models.Customer.objects.get(pk=subscription.customer.pk)
    return sync_subscription_from_stripe_data(customer, sub)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import datetime
import decimal
import sys

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import models
from django.test import TestCase
from django.utils import timezone

from mock import call, patch

from ..models import (
    Account,
    BankAccount,
    Card,
    Charge,
    Coupon,
    Customer,
    Event,
    EventProcessingException,
    Invoice,
    InvoiceItem,
    Plan,
    Subscription,
    Transfer,
    UserAccount
)

try:
    _str = unicode
except NameError:
    _str = str

PY2 = sys.version_info[0] == 2


class ModelTests(TestCase):

    def test_plan_str_and_repr(self):
        p = Plan(amount=decimal.Decimal("5"), name="My Plan", interval="monthly", interval_count=1)
        self.assertTrue(p.name in _str(p))
        self.assertEquals(repr(p), "Plan(pk=None, name={p.name!r}, amount=Decimal('5'), currency={p.currency!r}, interval={p.interval!r}, interval_count=1, trial_period_days=None, stripe_id={p.stripe_id!r})".format(p=p))

    def test_plan_repr_unicode(self):
        p = Plan(amount=decimal.Decimal("5"), name=u"öre", interval="monthly", interval_count=1, stripe_id=u"öre")
        if PY2:
            self.assertEquals(repr(p), "Plan(pk=None, name=u'\\xf6re', amount=Decimal('5'), currency=u'', interval=u'monthly', interval_count=1, trial_period_days=None, stripe_id=u'\\xf6re')")
        else:
            self.assertEquals(repr(p), "Plan(pk=None, name='öre', amount=Decimal('5'), currency='', interval='monthly', interval_count=1, trial_period_days=None, stripe_id='öre')")

    def test_plan_str_usd(self):
        p = Plan(amount=decimal.Decimal("5"), name="My Plan", currency="usd", interval="monthly", interval_count=1)
        self.assertTrue(u"\u0024" in _str(p))

    def test_plan_str_jpy(self):
        p = Plan(amount=decimal.Decimal("5"), name="My Plan", currency="jpy", interval="monthly", interval_count=1)
        self.assertTrue(u"\u00a5" in _str(p))

    @patch("stripe.Plan.retrieve")
    def test_plan_stripe_plan(self, RetrieveMock):
        c = Plan(stripe_id="plan")
        self.assertEqual(c.stripe_plan, RetrieveMock.return_value)
        self.assertTrue(RetrieveMock.call_args_list, [
            call("plan", stripe_account=None)])

    @patch("stripe.Plan.retrieve")
    def test_plan_stripe_plan_with_account(self, RetrieveMock):
        c = Plan(stripe_id="plan", stripe_account=Account(stripe_id="acct_A"))
        self.assertEqual(c.stripe_plan, RetrieveMock.return_value)
        self.assertTrue(RetrieveMock.call_args_list, [
            call("plan", stripe_account="acct_A")])

    def test_plan_per_account(self):
        Plan.objects.create(stripe_id="plan", amount=decimal.Decimal("100"), interval="monthly", interval_count=1)
        account = Account.objects.create(stripe_id="acct_A")
        Plan.objects.create(stripe_id="plan", stripe_account=account, amount=decimal.Decimal("100"), interval="monthly", interval_count=1)
        self.assertEquals(Plan.objects.count(), 2)

    def test_event_processing_exception_str(self):
        e = EventProcessingException(data="hello", message="hi there", traceback="fake")
        self.assertTrue("Event=" in str(e))

    def test_event_str_and_repr(self):
        created_at = datetime.datetime.utcnow()
        created_at_iso = created_at.replace(microsecond=0).isoformat()
        e = Event(kind="customer.deleted", webhook_message={}, created_at=created_at)
        self.assertTrue("customer.deleted" in str(e))
        if PY2:
            self.assertEquals(repr(e), "Event(pk=None, kind=u'customer.deleted', customer=None, valid=None, created_at={!s}, stripe_id=u'')".format(
                created_at_iso))
        else:
            self.assertEquals(repr(e), "Event(pk=None, kind='customer.deleted', customer=None, valid=None, created_at={!s}, stripe_id='')".format(
                created_at_iso))

        e.stripe_id = "evt_X"
        e.customer = Customer()
        if PY2:
            self.assertEquals(repr(e), "Event(pk=None, kind=u'customer.deleted', customer={!r}, valid=None, created_at={!s}, stripe_id=u'evt_X')".format(
                e.customer, created_at_iso))
        else:
            self.assertEquals(repr(e), "Event(pk=None, kind='customer.deleted', customer={!r}, valid=None, created_at={!s}, stripe_id='evt_X')".format(
                e.customer, created_at_iso))

    def test_customer_str_and_repr(self):
        c = Customer()
        self.assertEquals(str(c), "No User(s)")
        if PY2:
            self.assertEquals(repr(c), "Customer(pk=None, stripe_id=u'')")
        else:
            self.assertEquals(repr(c), "Customer(pk=None, stripe_id='')")

    def test_customer_with_user_str_and_repr(self):
        User = get_user_model()
        c = Customer(user=User())
        self.assertEqual(str(c), "")
        if PY2:
            self.assertEqual(repr(c), "Customer(pk=None, user=<User: >, stripe_id=u'')")
        else:
            self.assertEqual(repr(c), "Customer(pk=None, user=<User: >, stripe_id='')")

    def test_customer_saved_without_users_str(self):
        c = Customer.objects.create()
        self.assertEqual(str(c), "No User(s)")
        c.stripe_id = "cu_XXX"
        self.assertEqual(str(c), "No User(s) (cu_XXX)")

    def test_connected_customer_str_and_repr(self):
        User = get_user_model()
        user = User.objects.create()
        account = Account.objects.create(stripe_id="acc_A")
        customer = Customer.objects.create(stripe_id="cus_A", stripe_account=account)
        UserAccount.objects.create(customer=customer, user=user, account=account)
        self.assertEqual(str(customer), "")
        if PY2:
            self.assertEqual(repr(customer), "Customer(pk={c.pk}, users=<User: >, stripe_id=u'cus_A')".format(c=customer))
        else:
            self.assertEqual(repr(customer), "Customer(pk={c.pk}, users=<User: >, stripe_id='cus_A')".format(c=customer))

    def test_charge_repr(self):
        charge = Charge()
        if PY2:
            self.assertEquals(repr(charge), "Charge(pk=None, customer=None, source=u'', amount=None, captured=None, paid=None, stripe_id=u'')")
        else:
            self.assertEquals(repr(charge), "Charge(pk=None, customer=None, source='', amount=None, captured=None, paid=None, stripe_id='')")

    def test_charge_str(self):
        charge = Charge()
        self.assertEquals(str(charge), "$0 (unpaid, uncaptured)")
        charge.stripe_id = "ch_XXX"
        charge.captured = True
        charge.paid = True
        charge.amount = decimal.Decimal(5)
        self.assertEquals(str(charge), "$5")
        charge.refunded = True
        self.assertEquals(str(charge), "$5 (refunded)")

    def test_charge_total_amount(self):
        charge = Charge()
        self.assertEquals(charge.total_amount, 0)
        charge.amount = decimal.Decimal(17)
        self.assertEquals(charge.total_amount, 17)
        charge.amount_refunded = decimal.Decimal(15.5)
        self.assertEquals(charge.total_amount, 1.5)

    def test_plan_display_invoiceitem(self):
        p = Plan(amount=decimal.Decimal("5"), name="My Plan", interval="monthly", interval_count=1)
        p.save()
        i = InvoiceItem(plan=p)
        self.assertEquals(i.plan_display(), "My Plan")

    def test_coupon_percent(self):
        c = Coupon(percent_off=25, duration="repeating", duration_in_months=3)
        self.assertEquals(str(c), "Coupon for 25% off, repeating")

    def test_coupon_absolute(self):
        c = Coupon(amount_off=decimal.Decimal(50.00), duration="once", currency="usd")
        self.assertEquals(str(c), "Coupon for $50, once")

    def test_model_table_name(self):
        self.assertEquals(Customer()._meta.db_table, "pinax_stripe_customer")

    def test_event_message(self):
        event = Event(validated_message={"foo": 1})
        self.assertEquals(event.validated_message, event.message)

    def test_invoice_status(self):
        self.assertEquals(Invoice(paid=True).status, "Paid")

    def test_invoice_status_not_paid(self):
        self.assertEquals(Invoice(paid=False).status, "Open")

    def test_subscription_repr(self):
        s = Subscription()
        if PY2:
            self.assertEquals(repr(s), "Subscription(pk=None, customer=None, plan=None, status=u'', stripe_id=u'')")
        else:
            self.assertEquals(repr(s), "Subscription(pk=None, customer=None, plan=None, status='', stripe_id='')")
        s.customer = Customer()
        s.plan = Plan()
        s.status = "active"
        s.stripe_id = "sub_X"
        if PY2:
            self.assertEquals(
                repr(s),
                "Subscription(pk=None, customer={o.customer!r}, plan={o.plan!r}, status=u'active', stripe_id=u'sub_X')".format(o=s))
        else:
            self.assertEquals(
                repr(s),
                "Subscription(pk=None, customer={o.customer!r}, plan={o.plan!r}, status='active', stripe_id='sub_X')".format(o=s))

    def test_subscription_total_amount(self):
        sub = Subscription(plan=Plan(name="Pro Plan", amount=decimal.Decimal("100")), quantity=2)
        self.assertEquals(sub.total_amount, decimal.Decimal("200"))

    def test_subscription_plan_display(self):
        sub = Subscription(plan=Plan(name="Pro Plan"))
        self.assertEquals(sub.plan_display(), "Pro Plan")

    def test_subscription_status_display(self):
        sub = Subscription(status="overly_active")
        self.assertEquals(sub.status_display(), "Overly Active")

    def test_subscription_delete(self):
        plan = Plan.objects.create(stripe_id="pro2", amount=decimal.Decimal("100"), interval="monthly", interval_count=1)
        customer = Customer.objects.create(stripe_id="foo")
        sub = Subscription.objects.create(customer=customer, status="trialing", start=timezone.now(), plan=plan, quantity=1, cancel_at_period_end=True, current_period_end=(timezone.now() - datetime.timedelta(days=2)))
        sub.delete()
        self.assertIsNone(sub.status)
        self.assertEquals(sub.quantity, 0)
        self.assertEquals(sub.amount, 0)

    def test_account_str_and_repr(self):
        a = Account()
        self.assertEquals(str(a), " - ")
        if PY2:
            self.assertEquals(repr(a), "Account(pk=None, display_name=u'', type=None, authorized=True, stripe_id=u'')")
        else:
            self.assertEquals(repr(a), "Account(pk=None, display_name='', type=None, authorized=True, stripe_id='')")
        a.stripe_id = "acct_X"
        self.assertEquals(str(a), " - acct_X")
        if PY2:
            self.assertEquals(repr(a), "Account(pk=None, display_name=u'', type=None, authorized=True, stripe_id=u'acct_X')")
        else:
            self.assertEquals(repr(a), "Account(pk=None, display_name='', type=None, authorized=True, stripe_id='acct_X')")
        a.display_name = "Display name"
        a.authorized = False
        self.assertEquals(str(a), "Display name - acct_X")
        if PY2:
            self.assertEquals(repr(a), "Account(pk=None, display_name=u'Display name', type=None, authorized=False, stripe_id=u'acct_X')")
        else:
            self.assertEquals(repr(a), "Account(pk=None, display_name='Display name', type=None, authorized=False, stripe_id='acct_X')")

    @patch("stripe.Subscription.retrieve")
    def test_subscription_stripe_subscription_with_connnect(self, RetrieveMock):
        a = Account(stripe_id="acc_X")
        c = Customer(stripe_id="cus_X", stripe_account=a)
        s = Subscription(stripe_id="sub_X", customer=c)
        s.stripe_subscription
        RetrieveMock.assert_called_once_with("sub_X", stripe_account="acc_X")

    def test_customer_required_fields(self):
        c = Customer(stripe_id="cus_A")
        c.full_clean()

    def test_user_account_validation(self):
        User = get_user_model()
        a = Account()
        ua = UserAccount(user=User(), account=a, customer=Customer(stripe_account=Account()))
        with self.assertRaises(ValidationError):
            ua.clean()

    def test_user_account_repr(self):
        User = get_user_model()
        ua = UserAccount(user=User(), account=Account(), customer=Customer())
        self.assertEquals(
            repr(ua),
            "UserAccount(pk=None, user=<User: >, account={o.account!r}, customer={o.customer!r})".format(
                o=ua))

    def test_card_repr(self):
        card = Card(exp_month=1, exp_year=2000)
        self.assertEquals(repr(card), "Card(pk=None, customer=None)")

        card.customer = Customer.objects.create()
        card.save()
        self.assertEquals(repr(card), "Card(pk={c.pk}, customer={c.customer!r})".format(c=card))

    def test_blank_with_null(self):
        import inspect
        import pinax.stripe.models

        clsmembers = inspect.getmembers(pinax.stripe.models, inspect.isclass)
        classes = [x[1] for x in clsmembers
                   if issubclass(x[1], models.Model)]

        for klass in classes[0:1]:
            for f in klass._meta.fields:
                if f.null:
                    self.assertTrue(f.blank, msg="%s.%s should be blank=True" % (klass.__name__, f.name))


class StripeObjectTests(TestCase):

    @patch("stripe.Charge.retrieve")
    def test_stripe_charge(self, RetrieveMock):
        Charge().stripe_charge
        self.assertTrue(RetrieveMock.called)

    @patch("stripe.Customer.retrieve")
    def test_stripe_customer(self, RetrieveMock):
        Customer().stripe_customer
        self.assertTrue(RetrieveMock.called)

    @patch("stripe.Invoice.retrieve")
    def test_stripe_invoice(self, RetrieveMock):
        Invoice().stripe_invoice
        self.assertTrue(RetrieveMock.called)

    @patch("stripe.Subscription.retrieve")
    def test_stripe_subscription(self, RetrieveMock):
        Subscription(stripe_id="sub_X", customer=Customer(stripe_id="foo")).stripe_subscription
        RetrieveMock.assert_called_once_with("sub_X", stripe_account=None)

    @patch("stripe.Transfer.retrieve")
    def test_stripe_transfer(self, RetrieveMock):
        Transfer(amount=10).stripe_transfer
        self.assertTrue(RetrieveMock.called)

    @patch("stripe.Account.retrieve")
    def test_stripe_bankaccount(self, RetrieveMock):
        BankAccount(account=Account(stripe_id="foo")).stripe_bankaccount
        self.assertTrue(RetrieveMock.return_value.external_accounts.retrieve.called)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-11-23 20:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pinax_stripe', '0011_auto_20171121_1648'),
    ]

    operations = [
        migrations.RenameField(
            model_name='account',
            old_name='transfer_schedule_weekly_anchor',
            new_name='payout_schedule_weekly_anchor',
        ),
        migrations.RenameField(
            model_name='account',
            old_name='transfer_statement_descriptor',
            new_name='payout_statement_descriptor',
        ),
        migrations.RenameField(
            model_name='account',
            old_name='transfers_enabled',
            new_name='payouts_enabled',
        ),
        migrations.RemoveField(
            model_name='account',
            name='transfer_schedule_delay_days',
        ),
        migrations.RemoveField(
            model_name='account',
            name='transfer_schedule_interval',
        ),
        migrations.RemoveField(
            model_name='account',
            name='transfer_schedule_monthly_anchor',
        ),
        migrations.AddField(
            model_name='account',
            name='payout_schedule_delay_days',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='account',
            name='payout_schedule_interval',
            field=models.CharField(blank=True, choices=[('Manual', 'manual'), ('Daily', 'daily'), ('Weekly', 'weekly'), ('Monthly', 'monthly')], max_length=7, null=True),
        ),
        migrations.AddField(
            model_name='account',
            name='payout_schedule_monthly_anchor',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
    ]



from datetime import timedelta

from django.utils import timezone

from ..hooks import DefaultHookSet


class TestHookSet(DefaultHookSet):

    def adjust_subscription_quantity(self, customer, plan, quantity):
        """
        Given a customer, plan, and quantity, when calling Customer.subscribe
        you have the opportunity to override the quantity that was specified.

        Previously this was handled in the setting `PAYMENTS_PLAN_QUANTITY_CALLBACK`
        and was only passed a customer object.
        """
        return quantity or 4

    def trial_period(self, user, plan):
        """
        Given a user and plan, return an end date for a trial period, or None
        for no trial period.

        Was previously in the setting `TRIAL_PERIOD_FOR_USER_CALLBACK`
        """
        if plan is not None:
            return timezone.now() + timedelta(days=3)



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2017-05-09 17:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pinax_stripe', '0007_auto_20170108_1202'),
    ]

    operations = [
        migrations.AddField(
            model_name='invoice',
            name='tax',
            field=models.DecimalField(decimal_places=2, max_digits=9, null=True),
        ),
        migrations.AddField(
            model_name='invoice',
            name='tax_percent',
            field=models.DecimalField(decimal_places=2, max_digits=9, null=True),
        ),
    ]



from django.core.management.base import BaseCommand

from ...actions import charges


class Command(BaseCommand):

    help = "Check for newly available Charges."

    def handle(self, *args, **options):
        charges.update_charge_availability()



import decimal

from django.contrib.auth import get_user_model
from django.core import management
from django.test import TestCase

from mock import patch
from stripe.error import InvalidRequestError

from ..models import Coupon, Customer, Plan


class CommandTests(TestCase):

    def setUp(self):
        User = get_user_model()
        self.user = User.objects.create_user(username="patrick")

    @patch("stripe.Customer.retrieve")
    @patch("stripe.Customer.create")
    def test_init_customer_creates_customer(self, CreateMock, RetrieveMock):
        CreateMock.return_value = dict(
            account_balance=0,
            delinquent=False,
            default_source="card_178Zqj2eZvKYlo2Cr2fUZZz7",
            currency="usd",
            id="cus_XXXXX",
            sources=dict(data=[]),
            subscriptions=dict(data=[]),
        )
        management.call_command("init_customers")
        customer = Customer.objects.get(user=self.user)
        self.assertEquals(customer.stripe_id, "cus_XXXXX")

    @patch("stripe.Plan.all")
    @patch("stripe.Plan.auto_paging_iter", create=True, side_effect=AttributeError)
    def test_plans_create_deprecated(self, PlanAutoPagerMock, PlanAllMock):
        PlanAllMock().data = [{
            "id": "entry-monthly",
            "amount": 954,
            "interval": "monthly",
            "interval_count": 1,
            "currency": None,
            "statement_descriptor": None,
            "trial_period_days": None,
            "name": "Pro",
            "metadata": {}
        }]
        management.call_command("sync_plans")
        self.assertEquals(Plan.objects.count(), 1)
        self.assertEquals(Plan.objects.all()[0].stripe_id, "entry-monthly")
        self.assertEquals(Plan.objects.all()[0].amount, decimal.Decimal("9.54"))

    @patch("stripe.Plan.auto_paging_iter", create=True)
    def test_plans_create(self, PlanAutoPagerMock):
        PlanAutoPagerMock.return_value = [{
            "id": "entry-monthly",
            "amount": 954,
            "interval": "monthly",
            "interval_count": 1,
            "currency": None,
            "statement_descriptor": None,
            "trial_period_days": None,
            "name": "Pro",
            "metadata": {}
        }]
        management.call_command("sync_plans")
        self.assertEquals(Plan.objects.count(), 1)
        self.assertEquals(Plan.objects.all()[0].stripe_id, "entry-monthly")
        self.assertEquals(Plan.objects.all()[0].amount, decimal.Decimal("9.54"))

    @patch("stripe.Coupon.auto_paging_iter", create=True)
    def test_coupons_create(self, CouponAutoPagerMock):
        CouponAutoPagerMock.return_value = [{
            "id": "test-coupon",
            "object": "coupon",
            "amount_off": None,
            "created": 1482132502,
            "currency": "aud",
            "duration": "repeating",
            "duration_in_months": 3,
            "livemode": False,
            "max_redemptions": None,
            "metadata": {
            },
            "percent_off": 25,
            "redeem_by": None,
            "times_redeemed": 0,
            "valid": True
        }]
        management.call_command("sync_coupons")
        self.assertEquals(Coupon.objects.count(), 1)
        self.assertEquals(Coupon.objects.all()[0].stripe_id, "test-coupon")
        self.assertEquals(Coupon.objects.all()[0].percent_off, 25)

    @patch("stripe.Coupon.all")
    @patch("stripe.Coupon.auto_paging_iter", create=True, side_effect=AttributeError)
    def test_coupons_create_deprecated(self, CouponAutoPagerMock, CouponAllMock):
        CouponAllMock().data = [{
            "id": "test-coupon",
            "object": "coupon",
            "amount_off": None,
            "created": 1482132502,
            "currency": "aud",
            "duration": "repeating",
            "duration_in_months": 3,
            "livemode": False,
            "max_redemptions": None,
            "metadata": {
            },
            "percent_off": 25,
            "redeem_by": None,
            "times_redeemed": 0,
            "valid": True
        }]
        management.call_command("sync_coupons")
        self.assertEquals(Coupon.objects.count(), 1)
        self.assertEquals(Coupon.objects.all()[0].stripe_id, "test-coupon")
        self.assertEquals(Coupon.objects.all()[0].percent_off, 25)

    @patch("stripe.Customer.retrieve")
    @patch("pinax.stripe.actions.customers.sync_customer")
    @patch("pinax.stripe.actions.invoices.sync_invoices_for_customer")
    @patch("pinax.stripe.actions.charges.sync_charges_for_customer")
    def test_sync_customers(self, SyncChargesMock, SyncInvoicesMock, SyncMock, RetrieveMock):
        user2 = get_user_model().objects.create_user(username="thomas")
        get_user_model().objects.create_user(username="altman")
        Customer.objects.create(stripe_id="cus_XXXXX", user=self.user)
        Customer.objects.create(stripe_id="cus_YYYYY", user=user2)
        management.call_command("sync_customers")
        self.assertEqual(SyncChargesMock.call_count, 2)
        self.assertEqual(SyncInvoicesMock.call_count, 2)
        self.assertEqual(SyncMock.call_count, 2)

    @patch("stripe.Customer.retrieve")
    @patch("pinax.stripe.actions.customers.sync_customer")
    @patch("pinax.stripe.actions.invoices.sync_invoices_for_customer")
    @patch("pinax.stripe.actions.charges.sync_charges_for_customer")
    def test_sync_customers_with_test_customer(self, SyncChargesMock, SyncInvoicesMock, SyncMock, RetrieveMock):
        user2 = get_user_model().objects.create_user(username="thomas")
        get_user_model().objects.create_user(username="altman")
        Customer.objects.create(stripe_id="cus_XXXXX", user=self.user)
        Customer.objects.create(stripe_id="cus_YYYYY", user=user2)

        SyncMock.side_effect = InvalidRequestError("Unknown customer", None, http_status=404)

        management.call_command("sync_customers")
        self.assertEqual(SyncChargesMock.call_count, 0)
        self.assertEqual(SyncInvoicesMock.call_count, 0)
        self.assertEqual(SyncMock.call_count, 2)

    @patch("stripe.Customer.retrieve")
    @patch("pinax.stripe.actions.customers.sync_customer")
    @patch("pinax.stripe.actions.invoices.sync_invoices_for_customer")
    @patch("pinax.stripe.actions.charges.sync_charges_for_customer")
    def test_sync_customers_with_test_customer_unknown_error(self, SyncChargesMock, SyncInvoicesMock, SyncMock, RetrieveMock):
        user2 = get_user_model().objects.create_user(username="thomas")
        get_user_model().objects.create_user(username="altman")
        Customer.objects.create(stripe_id="cus_XXXXX", user=self.user)
        Customer.objects.create(stripe_id="cus_YYYYY", user=user2)

        SyncMock.side_effect = InvalidRequestError("Unknown error", None, http_status=500)

        with self.assertRaises(InvalidRequestError):
            management.call_command("sync_customers")
        self.assertEqual(SyncChargesMock.call_count, 0)
        self.assertEqual(SyncInvoicesMock.call_count, 0)
        self.assertEqual(SyncMock.call_count, 1)

    @patch("stripe.Customer.retrieve")
    @patch("pinax.stripe.actions.customers.sync_customer")
    @patch("pinax.stripe.actions.invoices.sync_invoices_for_customer")
    @patch("pinax.stripe.actions.charges.sync_charges_for_customer")
    def test_sync_customers_with_unicode_username(self, SyncChargesMock, SyncInvoicesMock, SyncMock, RetrieveMock):
        user2 = get_user_model().objects.create_user(username=u"tom\xe1s")
        Customer.objects.create(stripe_id="cus_YYYYY", user=user2)
        management.call_command("sync_customers")
        self.assertEqual(SyncChargesMock.call_count, 1)
        self.assertEqual(SyncInvoicesMock.call_count, 1)
        self.assertEqual(SyncMock.call_count, 1)

    @patch("stripe.Customer.retrieve")
    @patch("pinax.stripe.actions.invoices.sync_invoices_for_customer")
    @patch("pinax.stripe.actions.charges.sync_charges_for_customer")
    def test_sync_customers_with_remotely_purged_customer(self, SyncChargesMock, SyncInvoicesMock, RetrieveMock):
        customer = Customer.objects.create(
            user=self.user,
            stripe_id="cus_XXXXX"
        )

        RetrieveMock.return_value = dict(
            deleted=True
        )

        management.call_command("sync_customers")
        self.assertIsNone(Customer.objects.get(stripe_id=customer.stripe_id).user)
        self.assertIsNotNone(Customer.objects.get(stripe_id=customer.stripe_id).date_purged)
        self.assertEqual(SyncChargesMock.call_count, 0)
        self.assertEqual(SyncInvoicesMock.call_count, 0)

    @patch("pinax.stripe.actions.charges.update_charge_availability")
    def test_update_charge_availability(self, UpdateChargeMock):
        management.call_command("update_charge_availability")
        self.assertEqual(UpdateChargeMock.call_count, 1)



import os

import django

DEBUG = True
USE_TZ = True
TIME_ZONE = "UTC"
DATABASES = {
    "default": {
        "ENGINE": os.environ.get("PINAX_STRIPE_DATABASE_ENGINE", "django.db.backends.sqlite3"),
        "HOST": os.environ.get("PINAX_STRIPE_DATABASE_HOST", "127.0.0.1"),
        "NAME": os.environ.get("PINAX_STRIPE_DATABASE_NAME", "pinax_stripe"),
        "USER": os.environ.get("PINAX_STRIPE_DATABASE_USER", ""),
    }
}
MIDDLEWARE = [  # from 2.0 onwards, only MIDDLEWARE is used
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]
if django.VERSION < (1, 10):
    MIDDLEWARE_CLASSES = MIDDLEWARE
ROOT_URLCONF = "pinax.stripe.tests.urls"
INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "jsonfield",
    "pinax.stripe",
]
SITE_ID = 1
PINAX_STRIPE_PUBLIC_KEY = ""
PINAX_STRIPE_SECRET_KEY = "sk_test_01234567890123456789abcd"
PINAX_STRIPE_SUBSCRIPTION_REQUIRED_EXCEPTION_URLS = ["pinax_stripe_subscription_create"]
PINAX_STRIPE_SUBSCRIPTION_REQUIRED_REDIRECT = "pinax_stripe_subscription_create"
PINAX_STRIPE_HOOKSET = "pinax.stripe.tests.hooks.TestHookSet"
TEMPLATES = [{
    "BACKEND": "django.template.backends.django.DjangoTemplates",
    "DIRS": [
        "pinax/stripe/tests/templates"
    ],
    "APP_DIRS": True,
    "OPTIONS": {
        "debug": True,
        "context_processors": [
            "django.contrib.auth.context_processors.auth",
            "django.template.context_processors.debug",
            "django.template.context_processors.i18n",
            "django.template.context_processors.media",
            "django.template.context_processors.static",
            "django.template.context_processors.tz",
            "django.template.context_processors.request",
        ],
    },
}]
SECRET_KEY = "pinax-stripe-secret-key"



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-10-06 08:45
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pinax_stripe', '0004_plan_metadata'),
    ]

    operations = [
        migrations.AlterField(
            model_name='card',
            name='country',
            field=models.CharField(blank=True, max_length=2),
        ),
    ]



import datetime

from django.contrib.auth import get_user_model
from django.db import connection
from django.test import Client, RequestFactory, SimpleTestCase, TestCase
from django.test.utils import CaptureQueriesContext
from django.utils import timezone

from ..models import Account, Customer, Invoice, Plan, Subscription

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


User = get_user_model()


class AdminTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        super(AdminTestCase, cls).setUpClass()

        # create customers and current subscription records
        period_start = datetime.datetime(2013, 4, 1, tzinfo=timezone.utc)
        period_end = datetime.datetime(2013, 4, 30, tzinfo=timezone.utc)
        start = datetime.datetime(2013, 1, 1, tzinfo=timezone.utc)
        cls.plan = Plan.objects.create(
            stripe_id="p1",
            amount=10,
            currency="usd",
            interval="monthly",
            interval_count=1,
            name="Pro"
        )
        cls.plan2 = Plan.objects.create(
            stripe_id="p2",
            amount=5,
            currency="usd",
            interval="monthly",
            interval_count=1,
            name="Light"
        )
        for i in range(10):
            customer = Customer.objects.create(
                user=User.objects.create_user(username="patrick{0}".format(i)),
                stripe_id="cus_xxxxxxxxxxxxxx{0}".format(i)
            )
            Subscription.objects.create(
                stripe_id="sub_{}".format(i),
                customer=customer,
                plan=cls.plan,
                current_period_start=period_start,
                current_period_end=period_end,
                status="active",
                start=start,
                quantity=1
            )
        customer = Customer.objects.create(
            user=User.objects.create_user(username="patrick{0}".format(11)),
            stripe_id="cus_xxxxxxxxxxxxxx{0}".format(11)
        )
        Subscription.objects.create(
            stripe_id="sub_{}".format(11),
            customer=customer,
            plan=cls.plan,
            current_period_start=period_start,
            current_period_end=period_end,
            status="canceled",
            canceled_at=period_end,
            start=start,
            quantity=1
        )
        cls.customer = Customer.objects.create(
            user=User.objects.create_user(username="patrick{0}".format(12)),
            stripe_id="cus_xxxxxxxxxxxxxx{0}".format(12)
        )
        Subscription.objects.create(
            stripe_id="sub_{}".format(12),
            customer=customer,
            plan=cls.plan2,
            current_period_start=period_start,
            current_period_end=period_end,
            status="active",
            start=start,
            quantity=1
        )
        Invoice.objects.create(
            customer=customer,
            date=timezone.now(),
            amount_due=100,
            subtotal=100,
            total=100,
            period_end=period_end,
            period_start=period_start
        )
        cls.user = User.objects.create_superuser(
            username="admin", email="admin@test.com", password="admin")
        cls.account = Account.objects.create(stripe_id="acc_abcd")
        cls.client = Client()

    def setUp(self):
        try:
            self.client.force_login(self.user)
        except AttributeError:
            # Django 1.8
            self.client.login(username="admin", password="admin")

    def test_readonly_change_form(self):
        url = reverse("admin:pinax_stripe_customer_change", args=(self.customer.pk,))
        response = self.client.get(url)
        self.assertNotContains(response, "submit-row")

        response = self.client.post(url)
        self.assertEqual(response.status_code, 403)

    def test_customer_admin(self):
        """Make sure we get good responses for all filter options"""
        url = reverse("admin:pinax_stripe_customer_changelist")

        response = self.client.get(url + "?sub_status=active")
        self.assertEqual(response.status_code, 200)

        response = self.client.get(url + "?sub_status=cancelled")
        self.assertEqual(response.status_code, 200)

        response = self.client.get(url + "?sub_status=none")
        self.assertEqual(response.status_code, 200)

        response = self.client.get(url + "?has_card=yes")
        self.assertEqual(response.status_code, 200)

        response = self.client.get(url + "?has_card=no")
        self.assertEqual(response.status_code, 200)

    def test_customer_admin_prefetch(self):
        url = reverse("admin:pinax_stripe_customer_changelist")

        with CaptureQueriesContext(connection) as captured:
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)

        Customer.objects.create(
            user=User.objects.create_user(username="patrick{0}".format(13)),
            stripe_id="cus_xxxxxxxxxxxxxx{0}".format(13)
        )
        with self.assertNumQueries(len(captured)):
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)

    def test_invoice_admin(self):
        url = reverse("admin:pinax_stripe_invoice_changelist")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        response = self.client.get(url + "?has_card=no")
        self.assertEqual(response.status_code, 200)

        response = self.client.get(url + "?has_card=yes")
        self.assertEqual(response.status_code, 200)

    def test_plan_admin(self):
        url = reverse("admin:pinax_stripe_plan_changelist")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_charge_admin(self):
        url = reverse("admin:pinax_stripe_charge_changelist")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_account_filter(self):
        url = reverse("admin:pinax_stripe_customer_changelist")
        response = self.client.get(url + "?stripe_account={}".format(self.account.pk))
        self.assertEqual(response.status_code, 200)
        response = self.client.get(url + "?stripe_account=none")
        self.assertEqual(response.status_code, 200)

    @classmethod
    def get_changelist(cls, model_class, data=None):
        from django.contrib.admin.sites import AdminSite
        from django.utils.module_loading import import_string

        admin_class = import_string("pinax.stripe.admin.{}Admin".format(
            model_class.__name__))

        ma = admin_class(model_class, AdminSite())

        info = ma.model._meta.app_label, ma.model._meta.model_name
        url = reverse("admin:%s_%s_changelist" % info)
        request = RequestFactory().get(url, data=data)
        request.user = cls.user
        return ma.changelist_view(request).context_data["cl"]

    def test_account_search(self):
        cl = self.get_changelist(Account)
        self.assertEqual(list(cl.queryset), [self.account])

        cl = self.get_changelist(Account, {"q": "acc_doesnotexist"})
        self.assertEqual(list(cl.queryset), [])


class AdminSimpleTestCase(SimpleTestCase):

    def test_customer_user_without_user(self):
        from ..admin import customer_user

        class CustomerWithoutUser(object):
            user = None

        class Obj(object):
            customer = CustomerWithoutUser()

        self.assertEqual(customer_user(Obj()), "")



import decimal
import json

from django.contrib.auth import get_user_model
from django.dispatch import Signal
from django.test import TestCase
from django.test.client import Client

import six
import stripe
from mock import patch

from . import (
    PLAN_CREATED_TEST_DATA,
    TRANSFER_CREATED_TEST_DATA,
    TRANSFER_PENDING_TEST_DATA
)
from ..models import (
    Account,
    Customer,
    Event,
    EventProcessingException,
    Plan,
    Transfer
)
from ..webhooks import (
    AccountApplicationDeauthorizeWebhook,
    AccountExternalAccountCreatedWebhook,
    AccountUpdatedWebhook,
    ChargeCapturedWebhook,
    CustomerDeletedWebhook,
    CustomerSourceCreatedWebhook,
    CustomerSourceDeletedWebhook,
    CustomerSubscriptionCreatedWebhook,
    CustomerUpdatedWebhook,
    InvoiceCreatedWebhook,
    Webhook,
    registry
)

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


class WebhookRegistryTest(TestCase):

    def test_get_signal(self):
        signal = registry.get_signal("account.updated")
        self.assertTrue(isinstance(signal, Signal))

    def test_get_signal_keyerror(self):
        self.assertIsNone(registry.get_signal("not a webhook"))


class WebhookTests(TestCase):

    event_data = {
        "api_version": "2017-06-05",
        "created": 1348360173,
        "data": {
            "object": {
                "amount": 455,
                "currency": "usd",
                "date": 1348876800,
                "description": None,
                "id": "ach_XXXXXXXXXXXX",
                "object": "transfer",
                "other_transfers": [],
                "status": "pending",
                "livemode": True,
                "reversed": False,
                "summary": {
                    "adjustment_count": 0,
                    "adjustment_fee_details": [],
                    "adjustment_fees": 0,
                    "adjustment_gross": 0,
                    "charge_count": 1,
                    "charge_fee_details": [{
                        "amount": 45,
                        "application": None,
                        "currency": "usd",
                        "description": None,
                        "type": "stripe_fee"
                    }],
                    "charge_fees": 45,
                    "charge_gross": 500,
                    "collected_fee_count": 0,
                    "collected_fee_gross": 0,
                    "currency": "usd",
                    "net": 455,
                    "refund_count": 0,
                    "refund_fees": 0,
                    "refund_gross": 0,
                    "validation_count": 0,
                    "validation_fees": 0
                }
            }
        },
        "id": "evt_XXXXXXXXXXXXx",
        "livemode": True,
        "object": "event",
        "pending_webhooks": 1,
        "type": "transfer.created"
    }

    def test_webhook_init(self):
        event = Event(kind=None)
        webhook = Webhook(event)
        self.assertIsNone(webhook.name)

    @patch("stripe.Event.retrieve")
    @patch("stripe.Transfer.retrieve")
    def test_webhook_with_transfer_event(self, TransferMock, StripeEventMock):
        StripeEventMock.return_value.to_dict.return_value = self.event_data
        TransferMock.return_value = self.event_data["data"]["object"]
        msg = json.dumps(self.event_data)
        resp = Client().post(
            reverse("pinax_stripe_webhook"),
            six.u(msg),
            content_type="application/json"
        )
        self.assertEquals(resp.status_code, 200)
        self.assertTrue(Event.objects.filter(kind="transfer.created").exists())

    @patch("stripe.Event.retrieve")
    @patch("stripe.Transfer.retrieve")
    def test_webhook_associated_with_stripe_account(self, TransferMock, StripeEventMock):
        connect_event_data = self.event_data.copy()
        account = Account.objects.create(stripe_id="acc_XXX")
        connect_event_data["account"] = account.stripe_id
        StripeEventMock.return_value.to_dict.return_value = connect_event_data
        TransferMock.return_value = connect_event_data["data"]["object"]
        msg = json.dumps(connect_event_data)
        resp = Client().post(
            reverse("pinax_stripe_webhook"),
            six.u(msg),
            content_type="application/json"
        )
        self.assertEquals(resp.status_code, 200)
        self.assertTrue(Event.objects.filter(kind="transfer.created").exists())
        self.assertEqual(
            Event.objects.filter(kind="transfer.created").first().stripe_account,
            account
        )
        self.assertEquals(TransferMock.call_args_list, [
            [("ach_XXXXXXXXXXXX",), {"stripe_account": "acc_XXX"}],
        ])

    def test_webhook_duplicate_event(self):
        data = {"id": 123}
        Event.objects.create(stripe_id=123, livemode=True)
        msg = json.dumps(data)
        resp = Client().post(
            reverse("pinax_stripe_webhook"),
            six.u(msg),
            content_type="application/json"
        )
        self.assertEquals(resp.status_code, 200)
        dupe_event_exception = EventProcessingException.objects.get()
        self.assertEqual(dupe_event_exception.message, "Duplicate event record")
        self.assertEqual(str(dupe_event_exception.data), '{"id": 123}')

    def test_webhook_event_mismatch(self):
        event = Event(kind="account.updated")
        WH = registry.get("account.application.deauthorized")
        with self.assertRaises(Exception):
            WH(event)

    @patch("django.dispatch.Signal.send")
    def test_send_signal(self, SignalSendMock):
        event = Event(kind="account.application.deauthorized")
        WH = registry.get("account.application.deauthorized")
        WH(event).send_signal()
        self.assertTrue(SignalSendMock.called)

    def test_send_signal_not_sent(self):
        event = Event(kind="account.application.deauthorized")
        WH = registry.get("account.application.deauthorized")

        def signal_handler(sender, *args, **kwargs):
            self.fail("Should not have been called.")
        registry.get_signal("account.application.deauthorized").connect(signal_handler)
        webhook = WH(event)
        webhook.name = "mismatch name"  # Not sure how this ever happens due to the registry
        webhook.send_signal()

    @patch("pinax.stripe.actions.customers.link_customer")
    @patch("pinax.stripe.webhooks.Webhook.validate")
    @patch("pinax.stripe.webhooks.Webhook.process_webhook")
    def test_process_exception_is_logged(self, ProcessWebhookMock, ValidateMock, LinkMock):
        # note: we choose an event type for which we do no processing
        event = Event.objects.create(kind="account.external_account.created", webhook_message={}, valid=True, processed=False)
        ProcessWebhookMock.side_effect = stripe.StripeError("Message", "error")
        with self.assertRaises(stripe.StripeError):
            AccountExternalAccountCreatedWebhook(event).process()
        self.assertTrue(EventProcessingException.objects.filter(event=event).exists())

    @patch("pinax.stripe.actions.customers.link_customer")
    @patch("pinax.stripe.webhooks.Webhook.validate")
    @patch("pinax.stripe.webhooks.Webhook.process_webhook")
    def test_process_exception_is_logged_non_stripeerror(self, ProcessWebhookMock, ValidateMock, LinkMock):
        # note: we choose an event type for which we do no processing
        event = Event.objects.create(kind="account.external_account.created", webhook_message={}, valid=True, processed=False)
        ProcessWebhookMock.side_effect = Exception("generic exception")
        with self.assertRaises(Exception):
            AccountExternalAccountCreatedWebhook(event).process()
        self.assertTrue(EventProcessingException.objects.filter(event=event).exists())

    @patch("pinax.stripe.actions.customers.link_customer")
    @patch("pinax.stripe.webhooks.Webhook.validate")
    def test_process_return_none(self, ValidateMock, LinkMock):
        # note: we choose an event type for which we do no processing
        event = Event.objects.create(kind="account.external_account.created", webhook_message={}, valid=True, processed=False)
        self.assertIsNone(AccountExternalAccountCreatedWebhook(event).process())


class ChargeWebhookTest(TestCase):

    @patch("stripe.Charge.retrieve")
    @patch("pinax.stripe.actions.charges.sync_charge_from_stripe_data")
    def test_process_webhook(self, SyncMock, RetrieveMock):
        event = Event.objects.create(kind=ChargeCapturedWebhook.name, webhook_message={}, valid=True, processed=False)
        event.validated_message = dict(data=dict(object=dict(id=1)))
        ChargeCapturedWebhook(event).process_webhook()
        self.assertTrue(SyncMock.called)
        _, kwargs = RetrieveMock.call_args
        self.assertEquals(kwargs["expand"], ["balance_transaction"])
        self.assertEquals(kwargs["stripe_account"], None)

    @patch("stripe.Charge.retrieve")
    @patch("pinax.stripe.actions.charges.sync_charge_from_stripe_data")
    def test_process_webhook_connect(self, SyncMock, RetrieveMock):
        account = Account.objects.create(stripe_id="acc_A")
        event = Event.objects.create(kind=ChargeCapturedWebhook.name, webhook_message={}, valid=True, processed=False, stripe_account=account)
        event.validated_message = dict(data=dict(object=dict(id=1)))
        ChargeCapturedWebhook(event).process_webhook()
        self.assertTrue(SyncMock.called)
        _, kwargs = RetrieveMock.call_args
        self.assertEquals(kwargs["expand"], ["balance_transaction"])
        self.assertEquals(kwargs["stripe_account"], "acc_A")


class CustomerDeletedWebhookTest(TestCase):

    def test_process_webhook_without_linked_customer(self):
        event = Event.objects.create(kind=CustomerDeletedWebhook.name, webhook_message={}, valid=True, processed=False)
        CustomerDeletedWebhook(event).process_webhook()

    def test_process_webhook_with_linked_customer(self):
        User = get_user_model()
        customer = Customer.objects.create(user=User.objects.create())
        self.assertIsNotNone(customer.user)
        event = Event.objects.create(kind=CustomerDeletedWebhook.name, webhook_message={}, valid=True, processed=False, customer=customer)
        CustomerDeletedWebhook(event).process_webhook()
        self.assertIsNone(customer.user)


class CustomerUpdatedWebhookTest(TestCase):

    @patch("pinax.stripe.actions.customers.sync_customer")
    def test_process_webhook_without_customer(self, SyncMock):
        event = Event.objects.create(kind=CustomerUpdatedWebhook.name, webhook_message={}, valid=True, processed=False)
        CustomerUpdatedWebhook(event).process_webhook()
        self.assertEquals(SyncMock.call_count, 0)

    @patch("pinax.stripe.actions.customers.sync_customer")
    def test_process_webhook_without_customer_with_data(self, SyncMock):
        event = Event.objects.create(kind=CustomerUpdatedWebhook.name, webhook_message={}, valid=True, processed=False)
        obj = object()
        event.validated_message = dict(data=dict(object=obj))
        CustomerUpdatedWebhook(event).process_webhook()
        self.assertEquals(SyncMock.call_count, 0)

    @patch("pinax.stripe.actions.customers.sync_customer")
    def test_process_webhook_with_customer_with_data(self, SyncMock):
        customer = Customer.objects.create()
        event = Event.objects.create(kind=CustomerUpdatedWebhook.name, customer=customer, webhook_message={}, valid=True, processed=False)
        obj = object()
        event.validated_message = dict(data=dict(object=obj))
        CustomerUpdatedWebhook(event).process_webhook()
        self.assertEquals(SyncMock.call_count, 1)
        self.assertIs(SyncMock.call_args[0][0], customer)
        self.assertIs(SyncMock.call_args[0][1], obj)


class CustomerSourceCreatedWebhookTest(TestCase):

    @patch("pinax.stripe.actions.sources.sync_payment_source_from_stripe_data")
    def test_process_webhook(self, SyncMock):
        event = Event.objects.create(kind=CustomerSourceCreatedWebhook.name, webhook_message={}, valid=True, processed=False)
        event.validated_message = dict(data=dict(object=dict()))
        CustomerSourceCreatedWebhook(event).process_webhook()
        self.assertTrue(SyncMock.called)


class CustomerSourceDeletedWebhookTest(TestCase):

    @patch("pinax.stripe.actions.sources.delete_card_object")
    def test_process_webhook(self, SyncMock):
        event = Event.objects.create(kind=CustomerSourceDeletedWebhook.name, webhook_message={}, valid=True, processed=False)
        event.validated_message = dict(data=dict(object=dict(id=1)))
        CustomerSourceDeletedWebhook(event).process_webhook()
        self.assertTrue(SyncMock.called)


class PlanCreatedWebhookTest(TestCase):

    @patch("stripe.Event.retrieve")
    def test_plan_created(self, EventMock):
        ev = EventMock()
        ev.to_dict.return_value = PLAN_CREATED_TEST_DATA
        event = Event.objects.create(
            stripe_id=PLAN_CREATED_TEST_DATA["id"],
            kind="plan.created",
            livemode=True,
            webhook_message=PLAN_CREATED_TEST_DATA,
            validated_message=PLAN_CREATED_TEST_DATA,
            valid=True
        )
        registry.get(event.kind)(event).process()
        self.assertEquals(Plan.objects.all().count(), 1)


class PlanUpdatedWebhookTest(TestCase):

    @patch("stripe.Event.retrieve")
    def test_plan_created(self, EventMock):
        Plan.objects.create(
            stripe_id="gold1",
            name="Gold Plan",
            interval="month",
            interval_count=1,
            amount=decimal.Decimal("9.99")
        )
        ev = EventMock()
        ev.to_dict.return_value = PLAN_CREATED_TEST_DATA
        event = Event.objects.create(
            stripe_id=PLAN_CREATED_TEST_DATA["id"],
            kind="plan.updated",
            livemode=True,
            webhook_message=PLAN_CREATED_TEST_DATA,
            validated_message=PLAN_CREATED_TEST_DATA,
            valid=True
        )
        registry.get(event.kind)(event).process()
        plan = Plan.objects.get(stripe_id="gold1")
        self.assertEquals(plan.name, PLAN_CREATED_TEST_DATA["data"]["object"]["name"])


class CustomerSubscriptionCreatedWebhookTest(TestCase):

    @patch("pinax.stripe.actions.subscriptions.sync_subscription_from_stripe_data")
    @patch("pinax.stripe.actions.customers.sync_customer")
    def test_process_webhook(self, SyncMock, SubSyncMock):
        event = Event.objects.create(
            kind=CustomerSubscriptionCreatedWebhook.name,
            customer=Customer.objects.create(),
            validated_message={"data": {"object": {}}},
            valid=True,
            processed=False)
        CustomerSubscriptionCreatedWebhook(event).process_webhook()
        self.assertTrue(SyncMock.called)
        self.assertTrue(SubSyncMock.called)

    @patch("pinax.stripe.actions.subscriptions.sync_subscription_from_stripe_data")
    @patch("pinax.stripe.actions.customers.sync_customer")
    def test_process_webhook_no_customer(self, SyncMock, SubSyncMock):
        event = Event.objects.create(
            kind=CustomerSubscriptionCreatedWebhook.name,
            validated_message={"data": {"object": {}}},
            valid=True,
            processed=False)
        CustomerSubscriptionCreatedWebhook(event).process_webhook()
        self.assertFalse(SyncMock.called)
        self.assertTrue(SubSyncMock.called)


class CustomerSubscriptionUpdatedWebhookTest(TestCase):

    WEBHOOK_MESSAGE_DATA = {
        "object": {"livemode": False}
    }

    VALIDATED_MESSAGE_DATA = {
        "previous_attributes": {"days_until_due": 30, "billing": "send_invoice"},
        "object": {"livemode": False}
    }

    VALIDATED_MESSAGE_DATA_NOT_VALID = {
        "previous_attributes": {"days_until_due": 30, "billing": "send_invoice"},
        "object": {"livemode": True}
    }

    def test_is_event_valid_yes(self):
        self.assertTrue(Webhook.is_event_valid(self.WEBHOOK_MESSAGE_DATA, self.VALIDATED_MESSAGE_DATA))

    def test_is_event_valid_no(self):
        self.assertFalse(Webhook.is_event_valid(self.WEBHOOK_MESSAGE_DATA, self.VALIDATED_MESSAGE_DATA_NOT_VALID))


class InvoiceCreatedWebhookTest(TestCase):

    @patch("pinax.stripe.actions.invoices.sync_invoice_from_stripe_data")
    def test_process_webhook(self, SyncMock):
        event = Event.objects.create(kind=InvoiceCreatedWebhook.name, webhook_message={}, valid=True, processed=False)
        event.validated_message = dict(data=dict(object=dict(id=1)))
        InvoiceCreatedWebhook(event).process_webhook()
        self.assertTrue(SyncMock.called)


class TestTransferWebhooks(TestCase):

    @patch("stripe.Event.retrieve")
    @patch("stripe.Transfer.retrieve")
    def test_transfer_created(self, TransferMock, EventMock):
        ev = EventMock()
        ev.to_dict.return_value = TRANSFER_CREATED_TEST_DATA
        TransferMock.return_value = TRANSFER_CREATED_TEST_DATA["data"]["object"]
        event = Event.objects.create(
            stripe_id=TRANSFER_CREATED_TEST_DATA["id"],
            kind="transfer.created",
            livemode=True,
            webhook_message=TRANSFER_CREATED_TEST_DATA,
            validated_message=TRANSFER_CREATED_TEST_DATA,
            valid=True
        )
        registry.get(event.kind)(event).process()
        transfer = Transfer.objects.get(stripe_id="tr_XXXXXXXXXXXX")
        self.assertEquals(transfer.amount, decimal.Decimal("4.55"))
        self.assertEquals(transfer.status, "paid")

    @patch("stripe.Event.retrieve")
    @patch("stripe.Transfer.retrieve")
    def test_transfer_pending_create(self, TransferMock, EventMock):
        ev = EventMock()
        ev.to_dict.return_value = TRANSFER_PENDING_TEST_DATA
        TransferMock.return_value = TRANSFER_PENDING_TEST_DATA["data"]["object"]
        event = Event.objects.create(
            stripe_id=TRANSFER_PENDING_TEST_DATA["id"],
            kind="transfer.created",
            livemode=True,
            webhook_message=TRANSFER_PENDING_TEST_DATA,
            validated_message=TRANSFER_PENDING_TEST_DATA,
            valid=True
        )
        registry.get(event.kind)(event).process()
        transfer = Transfer.objects.get(stripe_id="tr_adlkj2l3kj23")
        self.assertEquals(transfer.amount, decimal.Decimal("9.41"))
        self.assertEquals(transfer.status, "pending")

    @patch("stripe.Event.retrieve")
    @patch("stripe.Transfer.retrieve")
    def test_transfer_paid_updates_existing_record(self, TransferMock, EventMock):
        ev = EventMock()
        ev.to_dict.return_value = TRANSFER_CREATED_TEST_DATA
        TransferMock.return_value = TRANSFER_CREATED_TEST_DATA["data"]["object"]
        event = Event.objects.create(
            stripe_id=TRANSFER_CREATED_TEST_DATA["id"],
            kind="transfer.created",
            livemode=True,
            webhook_message=TRANSFER_CREATED_TEST_DATA,
            validated_message=TRANSFER_CREATED_TEST_DATA,
            valid=True
        )
        registry.get(event.kind)(event).process()
        data = {
            "created": 1364658818,
            "data": {
                "object": {
                    "account": {
                        "bank_name": "BANK OF AMERICA, N.A.",
                        "country": "US",
                        "last4": "9999",
                        "object": "bank_account"
                    },
                    "amount": 455,
                    "currency": "usd",
                    "date": 1364601600,
                    "description": "STRIPE TRANSFER",
                    "fee": 0,
                    "fee_details": [],
                    "id": "tr_XXXXXXXXXXXX",
                    "livemode": True,
                    "object": "transfer",
                    "other_transfers": [],
                    "status": "paid",
                    "summary": {
                        "adjustment_count": 0,
                        "adjustment_fee_details": [],
                        "adjustment_fees": 0,
                        "adjustment_gross": 0,
                        "charge_count": 1,
                        "charge_fee_details": [{
                            "amount": 45,
                            "application": None,
                            "currency": "usd",
                            "description": None,
                            "type": "stripe_fee"
                        }],
                        "charge_fees": 45,
                        "charge_gross": 500,
                        "collected_fee_count": 0,
                        "collected_fee_gross": 0,
                        "collected_fee_refund_count": 0,
                        "collected_fee_refund_gross": 0,
                        "currency": "usd",
                        "net": 455,
                        "refund_count": 0,
                        "refund_fee_details": [],
                        "refund_fees": 0,
                        "refund_gross": 0,
                        "validation_count": 0,
                        "validation_fees": 0
                    },
                    "transactions": {
                        "count": 1,
                        "data": [{
                            "amount": 500,
                            "created": 1364064631,
                            "description": None,
                            "fee": 45,
                            "fee_details": [{
                                "amount": 45,
                                "application": None,
                                "currency": "usd",
                                "description": "Stripe processing fees",
                                "type": "stripe_fee"
                            }],
                            "id": "ch_XXXXXXXXXX",
                            "net": 455,
                            "type": "charge"
                        }],
                        "object": "list",
                        "url": "/v1/transfers/XX/transactions"
                    }
                }
            },
            "id": "evt_YYYYYYYYYYYY",
            "livemode": True,
            "object": "event",
            "pending_webhooks": 1,
            "type": "transfer.paid"
        }
        paid_event = Event.objects.create(
            stripe_id=data["id"],
            kind="transfer.paid",
            livemode=True,
            webhook_message=data,
            validated_message=data,
            valid=True
        )
        registry.get(paid_event.kind)(paid_event).process()
        transfer = Transfer.objects.get(stripe_id="tr_XXXXXXXXXXXX")
        self.assertEquals(transfer.status, "paid")


class AccountWebhookTest(TestCase):

    @classmethod
    def setUpClass(cls):
        super(AccountWebhookTest, cls).setUpClass()
        cls.account = Account.objects.create(stripe_id="acc_aa")

    @patch("stripe.Account.retrieve")
    @patch("pinax.stripe.actions.accounts.sync_account_from_stripe_data")
    def test_process_webhook(self, SyncMock, RetrieveMock):
        event = Event.objects.create(
            kind=AccountUpdatedWebhook.name,
            webhook_message={},
            valid=True,
            processed=False
        )
        event.validated_message = dict(data=dict(object=dict(id=1)))
        AccountUpdatedWebhook(event).process_webhook()
        self.assertTrue(SyncMock.called)

    @patch("stripe.Event.retrieve")
    def test_process_deauthorize(self, RetrieveMock):
        data = {"data": {"object": {"id": "evt_001"}},
                "account": self.account.stripe_id}
        event = Event.objects.create(
            kind=AccountApplicationDeauthorizeWebhook.name,
            webhook_message=data,
        )
        RetrieveMock.side_effect = stripe.error.PermissionError(
            "The provided key 'sk_test_********************abcd' does not have access to account 'acc_aa' (or that account does not exist). Application access may have been revoked.")
        AccountApplicationDeauthorizeWebhook(event).process()
        self.assertTrue(event.valid)
        self.assertTrue(event.processed)
        self.account.refresh_from_db()
        self.assertFalse(self.account.authorized)

    @patch("stripe.Event.retrieve")
    def test_process_deauthorize_fake_response(self, RetrieveMock):
        data = {"data": {"object": {"id": "evt_001"}},
                "account": self.account.stripe_id}
        event = Event.objects.create(
            kind=AccountApplicationDeauthorizeWebhook.name,
            webhook_message=data,
        )
        RetrieveMock.side_effect = stripe.error.PermissionError(
            "The provided key 'sk_test_********************ABCD' does not have access to account 'acc_aa' (or that account does not exist). Application access may have been revoked.")
        with self.assertRaises(stripe.error.PermissionError):
            AccountApplicationDeauthorizeWebhook(event).process()

    @patch("stripe.Event.retrieve")
    def test_process_deauthorize_with_delete_account(self, RetrieveMock):
        data = {"data": {"object": {"id": "evt_002"}},
                "account": "acct_bb"}
        event = Event.objects.create(
            kind=AccountApplicationDeauthorizeWebhook.name,
            webhook_message=data,
        )
        RetrieveMock.side_effect = stripe.error.PermissionError(
            "The provided key 'sk_test_********************abcd' does not have access to account 'acct_bb' (or that account does not exist). Application access may have been revoked.")
        AccountApplicationDeauthorizeWebhook(event).process()
        self.assertTrue(event.valid)
        self.assertTrue(event.processed)
        self.assertIsNone(event.stripe_account)

    @patch("stripe.Event.retrieve")
    def test_process_deauthorize_without_account(self, RetrieveMock):
        data = {"data": {"object": {"id": "evt_001"}}}
        event = Event.objects.create(
            kind=AccountApplicationDeauthorizeWebhook.name,
            webhook_message=data,
        )
        RetrieveMock.return_value.to_dict.return_value = data
        AccountApplicationDeauthorizeWebhook(event).process()
        self.assertTrue(event.valid)
        self.assertTrue(event.processed)
        self.account.refresh_from_db()
        self.assertTrue(self.account.authorized)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-11-21 16:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pinax_stripe', '0010_connect'),
    ]

    operations = [
        migrations.AlterField(
            model_name='plan',
            name='stripe_id',
            field=models.CharField(max_length=191),
        ),
        migrations.AlterUniqueTogether(
            name='plan',
            unique_together=set([('stripe_id', 'stripe_account')]),
        ),
    ]
